/********************************************************************************************
 *AKtron-IO-SPS v0.02
 *
 *Die Demo-Firmware AKtron-SPS veranschaulicht die Möglichkeit wie mit AKtron-IO eine
 *autark laufende SPS realisiert werden kann.
 *Weil AKtron-IO mehr Ein-/Ausgänge als das Arduino Mega2560 Board hat, wird auf
 *digitalRead() und digitalWrite() verzichtet und stattdessen Port-/Byteweise gelesen
 *und geschrieben.
 *Fuer das LOGO!8 Basismodul steht ein entsprechendes LOGO!Soft Comfort (LSC) Schaltprogramm
 *zur Verfügung.
 *
 *IDE:  Arduino 1.8.x
 *Mikrocontroller:  ATmega2560
 *Boardauswahl : Arduino AVR Boards -> Mega 2560
 ********************************************************************************************/
/********************************************************************************************
 * Die hier unter GNU General Public Lizenz v3 kostenlose und zur freien Verfuegung gestellte
 * OSS (Open Source Software) hat den Zweck dem Nutzer den Einstieg zu erleichtern und kann 
 * als Grundlage für eigene Erweiterungen, Projekte oder Schulungszwecke verwendet werden.
 * Bei den Softwareversionen 0.x handelt es sich um Demo-Versionen bei denen für Demozwecke
 * von „ausreichendem“ Funktionsumfang auszugehen ist.
 * Es besteht KEIN rechtlicher Anspruch auf Vollständigkeit und Fehlerfreiheit.
 *
 * Copyleft Andreas Kaestner https://aktron.de
 ********************************************************************************************/

// BOF preprocessor bug prevent - insert me on top of your arduino-code
#if 1
__asm volatile ("nop");
#endif

#define  EE_MySlaveID   152
#define  EE_Twi         153  // Twi (I2C) [0...1]

#define LED_ON  PORTJ|=128;
#define LED_OFF PORTJ&=127;

#include <avr/wdt.h>
#include <EEPROM.h>
#include "Adafruit_GFX.h"
#include "SH1106.h"

#define OLED_RESET 4
SH1106 display(OLED_RESET);

#define XPOS 0
#define YPOS 1

#if (SH1106_LCDHEIGHT != 64)
 #error("Height incorrect, please fix SH1106.h!");
#endif

String  sVersion = " v0.02";

unsigned long actGetIOTime, preGetIOTime;

byte    bBoardType = 0;
byte    bMySlaveID;
boolean xTwi = true;


byte bP, bX;          //  bP-Port 0...7, bX-Bit 0...7

// Ausgangs-Ports 4...7
byte    bQ[8]             = {0,0,0,0,0,0,0,0};  // aktueller Zustand Ausgang
byte    bQPre[8]          = {0,0,0,0,0,0,0,0};  // vorheriger Zustand Ausgang
byte    bQR[8]            = {0,0,0,0,0,0,0,0};  // Flanke positiv (rise)
byte    bQF[8]            = {0,0,0,0,0,0,0,0};  // Flanke negativ (fall)
byte    bQC[8]            = {0,0,0,0,0,0,0,0};  // PinChange

// Eingangs-Ports 0...3
byte    bI[8]             = {0,0,0,0,0,0,0,0};  // aktueller Zustand Eingang
byte    bIPre[8]          = {0,0,0,0,0,0,0,0};  // vorheriger Zustand Eingang
byte    bIR[8]            = {0,0,0,0,0,0,0,0};  // Flanke positiv (rise)
byte    bIF[8]            = {0,0,0,0,0,0,0,0};  // Flanke negativ (fall)
byte    bIC[8]            = {0,0,0,0,0,0,0,0};  // PinChange

byte    bReverse[8]       = {0,0,0,0,0,0,0,0};  // Port-Pins revers behandeln ( 0 = normal, 1 = revers bzw. gespiegelt)
byte    bInverse[8]       = {0,0,0,0,0,0,0,0};  // Port invers behandeln ( 0 = normal, 1 = invertiert)

// ************************************************************* 
//   ReverseByte() Eingangs Byte wird spielgelverkehrt
//   zurueck gegeben.
// *************************************************************
byte ReverseByte(byte in)
{
  byte out = 0;
  if (in & 0x01) out |= 0x80;
  if (in & 0x02) out |= 0x40;
  if (in & 0x04) out |= 0x20;
  if (in & 0x08) out |= 0x10;
  if (in & 0x10) out |= 0x08;
  if (in & 0x20) out |= 0x04;
  if (in & 0x40) out |= 0x02;
  if (in & 0x80) out |= 0x01;

  return out;
}
