
/*******************************************
 * !!! Hier verabschieden wir uns von digitalRead() und digitalWrite() !!!
 * 
 * Aber warum?
 * 
 * In AKtron-IO wird der gleiche Mikrocontroller verwendet, wie im Arduino Mega2560 Board
 * Beim Arduino Mega2560 stehen nur 69 IO-Pins zur Verfügung obwohl der Mikrocontroller 
 * insgesamt 86 IO-Pins hat. Mit dem Mega2560 Board koennen keine 64 IO-Pins für Schaltaufgaben
 * genutzt werden, ohne Kommunikationspins zu verwenden. Diese waeren dann nicht mehr nutzbar.
 * 
 * Bei den AKtron-IO Modulen wird diese Einschränkung umgangen
 *   - 64 Digitale IO-Pins stehen zur Verfuegung
 *   - 8  Pins für 4 serielle Schnittstellen (Rx/Tx)
 *   - 2  Pins für TWI/IIC
 *   - 3  Pins für SPI
 *   - 1  Pin OnBoard BlueLED
 *   - 1  Pin Abfrage Boardtyp
 *
 * Somit werden 79 Pins bei AKtron-IO benutzt.
 * Zudem werden die IO-Pins gruppiert und hier über Ports und Pins angesprochen
 * Das ist ausserhalb der Arduino IDE bei Mikrocontroller-Programmierung
 * und seit Jahrzehnten in der industriellen SPS-Programmierung üblich.
 * Die Aufrechterhaltung von digitalRead() und digitalWrite() haette einen deutlich groesseren 
 * Aufwand im Leiterplattendesign und eine hoehere Anzahl von Layer erfordert.
 */

/*******************************************
/* PortRead()
/*******************************************/
char PortRead(byte port, byte pin)
{
 byte mask = 1;
 
 if(port > 7 || pin > 7)
  return -1;

   mask <<= pin;
  
 if(port < 4)
 {// Eingang

   if((bI[port] & mask) > 0)
    return 1;
   else
    return 0;
 }
 else
 {// Ausgang
   if((bQ[port] & mask) > 0)
    return 1;
   else
    return 0; 
 }
 
}

/*******************************************
/* PortWrite()
/*******************************************/
char PortWrite(byte port, byte pin, boolean value) // Datentyp char wegen negativen Wert
{

  static byte x = 1;

  byte mask = 1;

  if(x)
  {
   Serial.println(port);
   Serial.println(pin);
   Serial.println((int)value);
   x=0;
  }
  
  if(port < 4 || port > 7 || pin > 7 || value > 1 || value < 0)
   return -1;

    mask <<= pin;

    if(value)
     return(bQ[port] |= mask);
    else
     return(bQ[port] &= (mask^0xFF));
 
}
