/*********************************************************************
This is a library for our Monochrome OLEDs based on SH1106 drivers

  Pick one up today in the adafruit shop!
  ------> http://www.adafruit.com/category/63_98

These displays use SPI to communicate, 4 or 5 pins are required to  
interface

Adafruit invests time and resources providing this open source code, 
please support Adafruit and open-source hardware by purchasing 
products from Adafruit!

Written by Limor Fried/Ladyada  for Adafruit Industries.  
BSD license, check license.txt for more information
All text above, and the splash screen below must be included in any redistribution
*********************************************************************/



#include "SH1106.h"


// the most basic function, set a single pixel
void SH1106::drawPixel(int16_t x, int16_t y, uint16_t color) {
  if ((x < 0) || (x >= width()) || (y < 0) || (y >= height()))
    return;

  // check rotation, move pixel around if necessary
  switch (getRotation()) {
  case 1:
    swap(x, y);
    x = WIDTH - x - 1;
    break;
  case 2:
    x = WIDTH - x - 1;
    y = HEIGHT - y - 1;
    break;
  case 3:
    swap(x, y);
    y = HEIGHT - y - 1;
    break;
  }  

  // x is which column
    switch (color) 
    {
      case WHITE:   buffer[x+ (y/8)*SH1106_LCDWIDTH] |=  (1 << (y&7)); break;
      case BLACK:   buffer[x+ (y/8)*SH1106_LCDWIDTH] &= ~(1 << (y&7)); break; 
      case INVERSE: buffer[x+ (y/8)*SH1106_LCDWIDTH] ^=  (1 << (y&7)); break; 
    }
    
}

// initializer for I2C - we only indicate the reset pin!
SH1106::SH1106(int8_t reset) :
Adafruit_GFX(SH1106_LCDWIDTH, SH1106_LCDHEIGHT) {
  sclk = dc = cs = sid = -1;
  rst = reset;
}

// ************************************************************* 
// Sequentielle "Multitasking"  TWI (I2C) Kommunikation
// *************************************************************
void SH1106::TwiBeginSimultan() {  

 delay(1); //Blocken verhindern
   
 switch(bTwiBeginStep)
 {
    // Init sequence for 128x64 OLED module
   case 1: 
    if(SH1106_commandTwi(SH1106_DISPLAYOFF) == 0)           // 0xAE
     bTwiBeginStep++;

    return;
   break; 

   case 2: 
    if(SH1106_commandTwi(SH1106_SETDISPLAYCLOCKDIV) == 0)            // 0xD5
     bTwiBeginStep++;

    return;
   break; 
   
   case 3: 
    if(SH1106_commandTwi(0x80) == 0)                                  // the suggested ratio 0x80
     bTwiBeginStep++;

    return;
   break; 

   case 4: 
    if(SH1106_commandTwi(SH1106_SETMULTIPLEX) == 0)                  // 0xA8
     bTwiBeginStep++;

    return;
   break; 

   case 5: 
    if(SH1106_commandTwi(0x3F) == 0)
     bTwiBeginStep++;

    return;
   break; 

   case 6: 
    if(SH1106_commandTwi(SH1106_SETDISPLAYOFFSET) == 0)              // 0xD3
     bTwiBeginStep++;

    return;
   break; 

   case 7: 
    if(SH1106_commandTwi(0x0) == 0)                                   // no offset
     bTwiBeginStep++;

    return;
   break; 

   case 8: 
    if(SH1106_commandTwi(SH1106_SETSTARTLINE | 0x0) == 0)            // line #0
     bTwiBeginStep++;

    return;
   break; 

   case 9: 
    if(SH1106_commandTwi(SH1106_CHARGEPUMP) == 0)                   // 0x8D
     bTwiBeginStep++;

    return;
   break; 

   case 10: 
    if(SH1106_commandTwi(0x14) == 0)    
     bTwiBeginStep++;

    return;
   break; 

   case 11: 
    if(SH1106_commandTwi(SH1106_MEMORYMODE) == 0)                    // 0x20
         bTwiBeginStep++;

    return;
   break; 

   case 12: 
    if(SH1106_commandTwi(0x00) == 0)                                 // 0x0 act like ks0108
     bTwiBeginStep++;

    return;
   break; 

   case 13: 
    if(SH1106_commandTwi(SH1106_SEGREMAP | 0x1) == 0)
     bTwiBeginStep++;

    return;
   break; 

   case 14: 
    if(SH1106_commandTwi(SH1106_COMSCANDEC) == 0)
     bTwiBeginStep++;

    return;
   break; 

   case 15: 
    if(SH1106_commandTwi(SH1106_SETCOMPINS) == 0)                    // 0xDA
     bTwiBeginStep++;

    return;
   break; 

   case 16: 
    if(SH1106_commandTwi(0x12) == 0)
     bTwiBeginStep++;

    return;
   break; 

   case 17: 
    if(SH1106_commandTwi(SH1106_SETCONTRAST) == 0)                   // 0x81
     bTwiBeginStep++;

    return;
   break; 

   case 18: 
    if(SH1106_commandTwi(0xCF) == 0)       
     bTwiBeginStep++;

    return;
   break; 

   case 19: 
    if(SH1106_commandTwi(SH1106_SETPRECHARGE) == 0)                  // 0xd9
     bTwiBeginStep++;

    return;
   break; 

   case 20: 
    if(SH1106_commandTwi(0xF1) == 0)   
          bTwiBeginStep++;

    return;
   break; 

   case 21: 
    if(SH1106_commandTwi(SH1106_SETVCOMDETECT) == 0)                 // 0xDB
     bTwiBeginStep++;

    return;
   break; 

   case 22: 
    if(SH1106_commandTwi(0x40) == 0)
     bTwiBeginStep++;

    return;
   break; 

   case 23: 
    if(SH1106_commandTwi(SH1106_DISPLAYALLON_RESUME) == 0)           // 0xA4
     bTwiBeginStep++;

    return;
   break; 

   case 24: 
    if(SH1106_commandTwi(SH1106_NORMALDISPLAY) == 0)                 // 0xA6
     bTwiBeginStep++;

    return;
   break; 

   case 25: 
    if(SH1106_commandTwi(SH1106_DISPLAYON) == 0)                    //--turn on oled panel
     {
      bTwiBeginStep = 1;
      xTwiBegin = false;
     }
      
   break; 
     
 }
 

}


void SH1106::invertDisplay(uint8_t i) {
  if (i) {
    SH1106_command(SH1106_INVERTDISPLAY);
  } else {
    SH1106_command(SH1106_NORMALDISPLAY);
  }
}


// ************************************************************* 
//
// *************************************************************
uint8_t SH1106::SH1106_commandTwi(uint8_t c) { 

 switch(bCommandTwi)
 {
   case 1:
    if(TWIStartSimultan(_i2caddr) == 0)
     bCommandTwi++;

     
    return bCommandTwi;
   break;

    case 2:

   //TWIWrite(0x00);
    if(TWIWriteSimultan(0x00) == 0)
      bCommandTwi++;
;
     
     return bCommandTwi;
    break;

    case 3:
    if(TWIWriteSimultan(c) == 0)
      bCommandTwi++;

     return bCommandTwi;
    break;

    case 4:
     TWIStop();
     bCommandTwi = 1;  // ende;-)
     return 0;
    break;
 } 
}

void SH1106::SH1106_command(uint8_t c) { 

    TWIStart(_i2caddr);
    TWIWrite(0x00);
    TWIWrite(c);
    TWIStop();
  
}

// ************************************************************* 
//
// *************************************************************
void SH1106::SH1106_data(uint8_t c) {

    TWIStart(_i2caddr);
    TWIWrite(0x40);
    TWIWrite(c);
    TWIStop();
  
}


// ************************************************************* 
// TWIStartSimultan
// *************************************************************
uint8_t SH1106::TWIStartSimultan(uint8_t bAdr)
{  
    switch(bStartSimultan)
    {

     case 1:
      TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
      bStartSimultan++;
      return bStartSimultan;
     break;

     case 2:
      TWDR = bAdr << 1;
      TWCR = (1<<TWINT)|(1<<TWEN);
      bStartSimultan = 1;
      return 0;
     break;
    }
}

// ************************************************************* 
// TWIWriteSimultan
// *************************************************************
uint8_t SH1106::TWIWriteSimultan(byte data)
{
  static byte b = 0;

    if((TWCR & (1<<TWINT)) == 0)
    {
      if(b++ == 64)
      {
        PORTJ |=  128;  //PIN J7 einschalten

        b = 0;
        
        Serial.print("TWI: 0x");
        Serial.println(data, HEX);
        delay(200);
        ReIniDisplay();
        PORTJ &=  127;  //PIN J7 ausschalten
        return 0;
      
      }
     return 1;
    }
    
    TWDR = data;
    TWCR = (1<<TWINT)|(1<<TWEN);
    return 0;

}

void SH1106::ReIniDisplay(void)
{
        TWCR = 0;
        bCommandTwi    = 1;
        bStartSimultan = 1;
        WriteToDisplayLoop = 0;
        pixel = 0;
        xData = 0;
        commandStep = 1; 
        pixelStep = 1;
        dataStep = 1; 
        
        TWIInit();        
        xTwiBegin = true;
  
}

// ************************************************************* 
// WriteToDisplay
// *************************************************************
void SH1106::WriteToDisplay(void) {

    while(WriteToDisplayLoop < 8) {

      switch(commandStep)
      {
        case 1:
         if(SH1106_commandTwi(SH1106_SETSTARTPAGE + WriteToDisplayLoop) == 0)
          commandStep++;
          
        return;
        break;
        
  // for some reason display is shifted by 2 columns
  // on 1.3" displays from ebay
       case 2:
        if(SH1106_commandTwi(SH1106_SETLOWCOLUMN + 2) == 0) // low column start address
         commandStep++;
         
        return;
        break;

        case 3:
        if(SH1106_commandTwi(SH1106_SETHIGHCOLUMN) == 0) // high column start address
         commandStep++;
         
        return;
        break;
      }
     
    while (pixel < SH1106_LCDWIDTH)
    {  
      switch(pixelStep)
      {             
       case 1:
        if(TWIStartSimultan(_i2caddr) == 0)
         pixelStep++;

        return;
       break;
       
       case 2:
       if(TWIWriteSimultan(0x40) == 0)
        pixelStep++;
        
        return;
       break;

       case 3:
       // input buffer doesn't accept all bytes at once
       while( xData<16) {
      
         if(TWIWriteSimultan(buffer[WriteToDisplayLoop * SH1106_LCDWIDTH + pixel])== 0)
         {
          xData++;
          pixel++;
          return;
         }
        }
  
       TWIStop();      
       xData = 0;
       pixelStep = 1;
       return;
      break;
      }     
    }

         commandStep = 1;

        if(pixel >= SH1106_LCDWIDTH)
        {

         pixel = 0;
         TWIStop();
 
        }
        
        WriteToDisplayLoop++;
        return;          
    }
    
 
    TWIStop(); 
    WriteToDisplayLoop = 0;
    return;
}



// clear everything
void SH1106::clearDisplay(void) {
  memset(buffer, 0, (SH1106_LCDWIDTH*SH1106_LCDHEIGHT/8));
}


void SH1106::ClearRegion(uint16_t from, uint16_t to)
{
  uint16_t b;
  register uint8_t *pBuf = buffer;

  pBuf--;
  
  for(b = from; b <= to; b++)
  {
    pBuf++;
   *pBuf = 0x00;
  }
}



void SH1106::drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color) {
  boolean bSwap = false;
  switch(rotation) { 
    case 0:
      // 0 degree rotation, do nothing
      break;
    case 1:
      // 90 degree rotation, swap x & y for rotation, then invert x
      bSwap = true;
      swap(x, y);
      x = WIDTH - x - 1;
      break;
    case 2:
      // 180 degree rotation, invert x and y - then shift y around for height.
      x = WIDTH - x - 1;
      y = HEIGHT - y - 1;
      x -= (w-1);
      break;
    case 3:
      // 270 degree rotation, swap x & y for rotation, then invert y  and adjust y for w (not to become h)
      bSwap = true;
      swap(x, y);
      y = HEIGHT - y - 1;
      y -= (w-1);
      break;
  }

  if(bSwap) { 
    drawFastVLineInternal(x, y, w, color);
  } else { 
    drawFastHLineInternal(x, y, w, color);
  }
}

void SH1106::drawFastHLineInternal(int16_t x, int16_t y, int16_t w, uint16_t color) {
  // Do bounds/limit checks
  if(y < 0 || y >= HEIGHT) { return; }

  // make sure we don't try to draw below 0
  if(x < 0) { 
    w += x;
    x = 0;
  }

  // make sure we don't go off the edge of the display
  if( (x + w) > WIDTH) { 
    w = (WIDTH - x);
  }

  // if our width is now negative, punt
  if(w <= 0) { return; }

  // set up the pointer for  movement through the buffer
  register uint8_t *pBuf = buffer;
  // adjust the buffer pointer for the current row
  pBuf += ((y/8) * SH1106_LCDWIDTH);
  // and offset x columns in
  pBuf += x;

  register uint8_t mask = 1 << (y&7);

  switch (color) 
  {
  case WHITE:         while(w--) { *pBuf++ |= mask; }; break;
    case BLACK: mask = ~mask;   while(w--) { *pBuf++ &= mask; }; break;
  case INVERSE:         while(w--) { *pBuf++ ^= mask; }; break;
  }
}

void SH1106::drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color) {
  bool bSwap = false;
  switch(rotation) { 
    case 0:
      break;
    case 1:
      // 90 degree rotation, swap x & y for rotation, then invert x and adjust x for h (now to become w)
      bSwap = true;
      swap(x, y);
      x = WIDTH - x - 1;
      x -= (h-1);
      break;
    case 2:
      // 180 degree rotation, invert x and y - then shift y around for height.
      x = WIDTH - x - 1;
      y = HEIGHT - y - 1;
      y -= (h-1);
      break;
    case 3:
      // 270 degree rotation, swap x & y for rotation, then invert y 
      bSwap = true;
      swap(x, y);
      y = HEIGHT - y - 1;
      break;
  }

  if(bSwap) { 
    drawFastHLineInternal(x, y, h, color);
  } else {
    drawFastVLineInternal(x, y, h, color);
  }
}


void SH1106::drawFastVLineInternal(int16_t x, int16_t __y, int16_t __h, uint16_t color) {

  // do nothing if we're off the left or right side of the screen
  if(x < 0 || x >= WIDTH) { return; }

  // maplayke sure we don't try to draw below 0
  if(__y < 0) { 
    // __y is negative, this will subtract enough from __h to account for __y being 0
    __h += __y;
    __y = 0;

  } 

  // make sure we don't go past the height of the display
  if( (__y + __h) > HEIGHT) { 
    __h = (HEIGHT - __y);
  }

  // if our height is now negative, punt 
  if(__h <= 0) { 
    return;
  }

  // this display doesn't need ints for coordinates, use local byte registers for faster juggling
  register uint8_t y = __y;
  register uint8_t h = __h;


  // set up the pointer for fast movement through the buffer
  register uint8_t *pBuf = buffer;
  // adjust the buffer pointer for the current row
  pBuf += ((y/8) * SH1106_LCDWIDTH);
  // and offset x columns in
  pBuf += x;

  // do the first partial byte, if necessary - this requires some masking
  register uint8_t mod = (y&7);
  if(mod) {
    // mask off the high n bits we want to set 
    mod = 8-mod;

    // note - lookup table results in a nearly 10% performance improvement in fill* functions
    // register uint8_t mask = ~(0xFF >> (mod));
    static uint8_t premask[8] = {0x00, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE };
    register uint8_t mask = premask[mod];

    // adjust the mask if we're not going to reach the end of this byte
    if( h < mod) { 
      mask &= (0XFF >> (mod-h));
    }

  switch (color) 
    {
    case WHITE:   *pBuf |=  mask;  break;
    case BLACK:   *pBuf &= ~mask;  break;
    case INVERSE: *pBuf ^=  mask;  break;
    }
  
    // fast exit if we're done here!
    if(h<mod) { return; }

    h -= mod;

    pBuf += SH1106_LCDWIDTH;
  }


  // write solid bytes while we can - effectively doing 8 rows at a time
  if(h >= 8) { 
    if (color == INVERSE)  {          // separate copy of the code so we don't impact performance of the black/white write version with an extra comparison per loop
      do  {
      *pBuf=~(*pBuf);

        // adjust the buffer forward 8 rows worth of data
        pBuf += SH1106_LCDWIDTH;

        // adjust h & y (there's got to be a faster way for me to do this, but this should still help a fair bit for now)
        h -= 8;
      } while(h >= 8);
      }
    else {
      // store a local value to work with 
      register uint8_t val = (color == WHITE) ? 255 : 0;

      do  {
        // write our value in
      *pBuf = val;

        // adjust the buffer forward 8 rows worth of data
        pBuf += SH1106_LCDWIDTH;

        // adjust h & y (there's got to be a faster way for me to do this, but this should still help a fair bit for now)
        h -= 8;
      } while(h >= 8);
      }
    }

  // now do the final partial byte, if necessary
  if(h) {
    mod = h & 7;
    // this time we want to mask the low bits of the byte, vs the high bits we did above
    // register uint8_t mask = (1 << mod) - 1;
    // note - lookup table results in a nearly 10% performance improvement in fill* functions
    static uint8_t postmask[8] = {0x00, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F };
    register uint8_t mask = postmask[mod];
    switch (color) 
    {
      case WHITE:   *pBuf |=  mask;  break;
      case BLACK:   *pBuf &= ~mask;  break;
      case INVERSE: *pBuf ^=  mask;  break;
    }
  }
}


// ************************************************************* 
//
// ************************************************************* 
void SH1106::TWIInit(void)
{
  // activate internal pullups for twi.
  digitalWrite(SDA, 1);
  digitalWrite(SCL, 1);

  // initialize twi prescaler and bit rate
  cbi(TWSR, TWPS0);
  cbi(TWSR, TWPS1);
  TWBR = ((F_CPU / TWI_FREQ) - 16) / 2;
  
    //enable TWI
    TWCR = (1<<TWEN);
}

// ************************************************************* 
//
// *************************************************************
void SH1106::TWIStart(uint8_t bAdr)
{
    TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
    while ((TWCR & (1<<TWINT)) == 0);

    TWDR = bAdr << 1;
    TWCR = (1<<TWINT)|(1<<TWEN);
    while ((TWCR & (1<<TWINT)) == 0);
}

// ************************************************************* 
// send stop signal
// *************************************************************
void SH1106::TWIStop(void)
{
    TWCR = (1<<TWINT)|(1<<TWSTO)|(1<<TWEN);
}

// ************************************************************* 
//
// *************************************************************
void SH1106::TWIWrite(uint8_t data)
{
  // PORTJ |=  128;  //PIN J7 einschalten
    TWDR = data;
    TWCR = (1<<TWINT)|(1<<TWEN);
    while ((TWCR & (1<<TWINT)) == 0);
  // PORTJ &=  127;  //PIN J7 ausschalten
}

// ************************************************************* 
//
// *************************************************************
uint8_t TWIReadACK(void)
{
    TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWEA);
    while ((TWCR & (1<<TWINT)) == 0);
    return TWDR;
}

// ************************************************************* 
//
// *************************************************************
//read byte with NACK
uint8_t TWIReadNACK(void)
{
    TWCR = (1<<TWINT)|(1<<TWEN);
    while ((TWCR & (1<<TWINT)) == 0);
    return TWDR;
}
// ************************************************************* 
//
// *************************************************************
uint8_t TWIGetStatus(void)
{
    uint8_t s;
    //mask status
    s = TWSR & 0xF8;
    return s;
}
